
#============================================================================
# Tangent-V specific Indexing parameter
#
# Kenny Davila
# - Created:  January 2018
#
#============================================================================

from AM_CommonTools.configuration.configuration import Configuration

class IndexParameters:
    # Similarity Metrics ... (Fixed)
    Metric_ProductBased = 0  # original
    Metric_MaximumLineOfSightSimilarity = 1  # MLS
    Metric_MinBased = 2  # Use min confidence for matching labels
    Metric_MinBasedMLS = 3  # MLS - using min confidence for weighted recall

    WildcardsWeight_Full = 0
    WildcardsWeight_Same = 1
    WildcardsWeight_Average = 2

    def __init__(self):
        # Main system parameters (and their default values)

        # Related to Graph Recognition
        self.Graph_KNN = 2
        self.Graph_TopClasses = 10
        self.Graph_MinScale = 0.333
        self.Graph_MaxScale = 3.0
        self.Graph_MaxEdgeDistance = 2.0

        # Related to pair generation
        self.Pairs_MinConfidence = 0.8
        self.Pairs_SmallExpressions = 0

        # Related to Raw Index Pair Matching
        self.IndexEntry_MinCosineSimilarity = 0.707107  # 0.0 (0.0 deg.), 0.707107 (45 deg.)
        self.IndexEntry_ScaleCosineSimilarity = False
        self.IndexEntry_MinSizeRatio = 0.0

        # Layer 1 - using GPU ...
        self.EntryMatching_UseTorch = False
        self.EntryMatching_MinTorchSize = 50000

        # End of Layer 1
        self.Top_Layer1_candidates = 100
        self.Layer1_ReturnRankings = False

        # Related to detailed matching process ...(mostly Layer 2)
        self.Match_MaxPathLength = 4
        self.Match_MaxSpatialScore = 3.0
        self.Match_MaxDisconnectedSpatialScore = 3.0
        self.Match_EarlyDeletion = True
        self.Match_MinGroupAreaOverlap = 0.5

        # Related to handling of wildcards
        self.Wildcards_Match = False
        self.Wildcards_WeightMode = 0

        # Similarity metrics (must be defined by the config)
        self.Layer1_SimilarityMetric = -1
        self.Layer2_SimilarityMetric = -1

        # Layer 2 - Step 2
        self.ConnPartComp_MaxImgInCandidates = 1000
        self.ConnPartComp_MaxImgOutCandidates = 1000
        self.ConnPartComp_MaxNeighbors = 1000

        # Layer 2
        # (By default, use one process only)
        self.Layer2_Workers = 1
        self.Layer2_MinParallelQuerySize = 10
        # ... optional step 3
        self.Layer2_JoinConnectedMatches = False
        # .... Optional step 5....
        self.Layer2_JoinDisconnectedMatches = False

        # Layer 3
        self.Layer3_PrecisionScore = 0

    @staticmethod
    def ReadFromConfiguration(config):
        assert isinstance(config, Configuration)

        # create default configuration
        params = IndexParameters()

        # Override default values using configuration file values...

        if config.contains("ENTRY_MIN_COS_SIMILARITY"):
            params.IndexEntry_MinCosineSimilarity = config.get_float("ENTRY_MIN_COS_SIMILARITY")
        if config.contains("ENTRY_SCALE_COS_SIMILARITY"):
            params.IndexEntry_ScaleCosineSimilarity = config.get_bool("ENTRY_SCALE_COS_SIMILARITY")
        if config.contains("ENTRY_MIN_SIZE_RATIO"):
            params.IndexEntry_MinSizeRatio = config.get_float("ENTRY_MIN_SIZE_RATIO")

        if config.contains("LAYER1_USE_PYTORCH"):
            params.EntryMatching_UseTorch = config.get_bool("LAYER1_USE_PYTORCH")
        if config.contains("LAYER1_MIN_PYTORCH_SIZE"):
            params.EntryMatching_MinTorchSize = config.get_int("LAYER1_MIN_PYTORCH_SIZE")

        if config.contains("GRAPH_KNN"):
            params.Graph_KNN = config.get_int("GRAPH_KNN")
        if config.contains("GRAPH_TOP_CLASSES"):
            params.Graph_TopClasses = config.get_int("GRAPH_TOP_CLASSES")
        if config.contains("GRAPH_MIN_SCALE"):
            params.Graph_MinScale = config.get("GRAPH_MIN_SCALE")
        if config.contains("GRAPH_MAX_SCALE"):
            params.Graph_MaxScale = config.get("GRAPH_MAX_SCALE")
        if config.contains("GRAPH_MAX_EDGE_DISTANCE"):
            params.Graph_MaxEdgeDistance = config.get_float("GRAPH_MAX_EDGE_DISTANCE")

        if config.contains("PAIRS_MIN_CONFIDENCE"):
            params.Pairs_MinConfidence = config.get_float("PAIRS_MIN_CONFIDENCE")
        if config.contains("PAIRS_SMALL_EXPRESSION"):
            params.Pairs_SmallExpressions = config.get_int("PAIRS_SMALL_EXPRESSION")

        if config.contains("MATCH_MAX_PATH_LENGTH"):
            params.Match_MaxPathLength = config.get_int("MATCH_MAX_PATH_LENGTH")
        if config.contains("MATCH_MAX_SPATIAL_SCORE"):
            params.Match_MaxSpatialScore = config.get_float("MATCH_MAX_SPATIAL_SCORE")
        if config.contains("MATCH_MAX_DISCONNECTED_SPATIAL_SCORE"):
            params.Match_MaxDisconnectedSpatialScore = config.get_float("MATCH_MAX_DISCONNECTED_SPATIAL_SCORE")
        if config.contains("MATCH_MIN_GROUP_AREA_OVERLAP"):
            params.Match_MinGroupAreaOverlap = config.get_float("MATCH_MIN_GROUP_AREA_OVERLAP")
        if config.contains("MATCH_EARLY_DELETION"):
            params.Match_EarlyDeletion = config.get_bool("MATCH_EARLY_DELETION")

        if config.contains("WILDCARDS_MATCH"):
            params.Wildcards_Match = config.get_bool("WILDCARDS_MATCH")
        if config.contains("WILDCARDS_WEIGHT_MODE"):
            params.Wildcards_WeightMode = config.get_int("WILDCARDS_WEIGHT_MODE")

        if config.contains("CONN_PART_COMP_MAX_IN_IMG_CANDIDATES"):
            params.ConnPartComp_MaxImgInCandidates = config.get_int("CONN_PART_COMP_MAX_IN_IMG_CANDIDATES")
        if config.contains("CONN_PART_COMP_MAX_OUT_IMG_CANDIDATES"):
            params.ConnPartComp_MaxImgOutCandidates = config.get_int("CONN_PART_COMP_MAX_OUT_IMG_CANDIDATES")
        if config.contains("CONN_PART_COMP_MAX_CAND_NEIGHBORS"):
            params.ConnPartComp_MaxNeighbors = config.get_int("CONN_PART_COMP_MAX_CAND_NEIGHBORS")

        if config.contains("LAYER1_SIMILARITY_METRIC"):
            params.Layer1_SimilarityMetric = config.get_int("LAYER1_SIMILARITY_METRIC")
        if config.contains("LAYER2_SIMILARITY_METRIC"):
            params.Layer2_SimilarityMetric = config.get_int("LAYER2_SIMILARITY_METRIC")

        if config.contains("TOP_LAYER1_CANDIDATES"):
            params.Top_Layer1_candidates = config.get_int("TOP_LAYER1_CANDIDATES")
        if config.contains("LAYER1_RETURN_INITIAL_RANKINGS"):
            params.Layer1_ReturnRankings = config.get_bool("LAYER1_RETURN_INITIAL_RANKINGS")

        if config.contains("LAYER2_WORKERS"):
            params.Layer2_Workers = config.get_int("LAYER2_WORKERS")

        if config.contains("LAYER2_MIN_PARALLEL_QUERY_SIZE"):
            params.Layer2_MinParallelQuerySize = config.get_int("LAYER2_MIN_PARALLEL_QUERY_SIZE")
        if config.contains("LAYER2_JOIN_CONNECTED_MATCHES"):
            params.Layer2_JoinConnectedMatches = config.get_bool("LAYER2_JOIN_CONNECTED_MATCHES")
        if config.contains("LAYER2_JOIN_DISCONNECTED_MATCHES"):
            params.Layer2_JoinDisconnectedMatches = config.get_bool("LAYER2_JOIN_DISCONNECTED_MATCHES")

        if config.contains("LAYER3_PRECISION_SCORE"):
            params.Layer3_PrecisionScore = config.get_int("LAYER3_PRECISION_SCORE")

        deprecated_params = ["MATCH_WILDCARDS", "SIMILARITY_METRIC"]
        for deprecated_param in deprecated_params:
            if config.contains(deprecated_param):
                raise Exception("Deprecated parameter name found: " + deprecated_param)

        return params