
import os
import re
import random

import cv2
import sympy

from TangentV.indexing.recognition_helper import RecognitionHelper

class LaTeX_Helper:
    RESymbolCommand = re.compile(r'\\[a-zA-Z]+symbol')
    REAndCommand = re.compile(r'\\and[a-zA-Z]*')
    RenderPreamble = None

    @staticmethod
    def replace_invalid_chars(latex_string):
        rep = {"&gt;": ">", "&lt;": "<", "≠": "\\neq ", "−": "-", "·": "\cdot ",
               "‹": "\langle ", "›": "\\rangle ", "⟨": "\langle ", "⟩": "\\rangle ", "〈": "\langle ", "〉": "\\rangle ",

               "∧": "\land ", "∨": "\lor ",
               "÷": "\div ","√": "\sqrt{~}",

               "∕": "/", "⁄": "/", '∖': "\\backslash ",

               "⊢": "\\vdash", "⊧": "\\vDash ", "⊥": "\\bot ",
               "≪": "\ll ", "≫": "\gg ",

               "⊕": "\oplus ", "⊗": "\otimes ",
               "∠": "\\angle ",

               "≤": "\leq ", "≥": "\geq ",
               "~": "\sim ", "≈": "\\approx ", "≡": "\equiv ", "≅": "\cong ",
               "≔": "\coloneqq ", "≐": "\doteq ",

               "‘": "`", "’": "'", "“": "``", "”": "''", "″": "''",  "´": "'",

               "¬": "\\neg ", "′": "'", "★": "\\bigstar ",
               "–": "-", "—": "-",
               "∞": "\infty ", "‖": "\parallel ", "±": "\pm ", "…": "\ldots ",
               "∇": "\\nabla ", "∆": "\\triangle ",

               "†": "\dagger ",

               "←": "\leftarrow ", "↑": "\\uparrow ","→": "\\rightarrow ", "↓": "\downarrow ",
               "⇐": "\Leftarrow ", "⇒": "\Rightarrow ",
               "⇔": "\Leftrightarrow ",
               "↔": "\leftrightarrow ", "↦": "\mapsto ",

               "°": "^{\circ} ", "∘": "\circ ",
               "̄": " ^{-}", "ˇ": "\\text{\\textasciicaron}",

               "£": "\\text{\\textlira}",

               "Ɛ": "E",
               
               "Đ": "D", "Ł": "\L ", "Ø": "\O ", "∅": "\oslash ",
               "đ": "d", "ħ": "h", "ł": "\l ", "ø": "\o ",

               "∂": "\partial ", "∫": "\int ",

               "∀": "\\forall ", "∃": "\exists ", "∪": "\cup ", "∩": "\cap ",
               "⊂": "\subset ", "⊆": "\subseteq ", "⊇": "\supseteq ", "⊈": "\\nsubseteq ",
               "∉": "\\notin ", "∈": "\in ",
               "⋐": "\Subset ", "⊋": "\\varsupsetneq ",

               "ℂ": "\mathbb{C} ", "ℍ": "\mathbb{H} ", "ℝ": "\mathbb{R} ", "ℤ": "\mathbb{Z} ",
               "𝖆": "a", "𝖌": "g", "𝖓": "n",
               "ℑ": "I", "𝔛": "x",
               "ℓ": "l",
               "α": "\\alpha ", "β": "\\beta ", "γ": "\gamma ", "δ": "\delta ", "ε": "\\varepsilon ",
               "ζ": "\zeta ", "η": "\\eta ", "θ": "\\theta ", "ι": "\iota ", "κ": "\kappa ",
               "λ": "\lambda ", "μ": "\mu ", "ν": "\\nu ", "ξ": "\\xi ", "ο": "o ",
               "π": "\pi ", "ρ": "\\rho ", "ς": "\\varsigma ", "σ": "\sigma ","τ": "\\tau ",
               "υ": "\\upsilon ", "φ": "\\varphi ", "χ": "\chi ", "ψ": "\psi ", "ω": "\omega ",
               "ϕ": "\phi ",

               "ί": "\\acute{\iota}", "ἰ": "\iota ",
               "ϵ": "\epsilon", "µ": "\mu ",
               "ώ": "\\acute{\omega}",

               "Λ": "\Lambda ",
               "Γ": "\Gamma ",
               "Δ": "\Delta ",
               "Ω": "\Omega ", "Ψ": "\Psi ", "Θ": "\\varTheta ",
               "Φ": "\Phi ", "Π": "\Pi ", "Ξ": "\Xi ",

               "Σ": "\Sigma ", "Τ": "T",
               "∑": "\sum ",

               "ϛ": "\sigma ", # incorrect, but couldn't find mapping for Stigma

               "ƒ": "f ", "§": "\S ",
               "ℒ": "L",
               "Α": "A", "Β": "B", "Η": "H", "Ι": "I", "Κ": "K", "Μ": "M",
               "Ν": "N", "Ο": "O", "Ρ": "P", "Χ": "X", "Υ": "Y",
               "ɡ": "g",

               "ḡ": "\\bar{\\textsl{g}} ",

               "ℕ": "\mathbb{N}", "ℙ": "\mathbb{P}", "ℚ": "\mathbb{Q}",

               "ℜ": "\mathfrak{R}",

               "𝖙": "\mathfrak{\\textbf{t}} ",

               "ṙ": "\dot{r}", "ẇ": "\dot{w}",

               "ȳ": "\\text{\={y}}",

               "á": "\\acute{a}", "é": "\\acute{e}", "í": "\\acute{i}", "ó": "\\acute{o}", "ú": "\\acute{u}",
               "ä": "\ddot{a}", "ë": "\ddot{e}", "ï": "\ddot{i}",  "ö": "\ddot{o}", "ü": "\ddot{u}",
               "â": "\hat{a}", "ê": "\hat{e}", "î": "\hat{i}", "ô": "\hat{o}", "û": "\hat{u}",

               "ŷ": "\hat{y}",
               "Â": "\hat{A}", "Ê": "\hat{E}", "Î": "\hat{I}", "Ô": "\hat{O}", "Û": "\hat{U}",
               "Ĥ": "\hat{H}", "Ŷ": "\hat{Y}",

               "ũ": "\\tilde{u}",

               "ℎ": "\hbar ", "ℏ": "\hbar ",

               "⁰": " ^{0}", "¹": " ^{1}", "²": " ^{2}", "³": " ^{3}",
               "⁻": " ^{-}",
               "₊": " _{+}", "₋": "_{-}",
               "₁": " _{1}", "₂": " _{2}",
               "ᴾ": " ^{P}", "ᵀ": " ^{T}",
               "ᵢ": " _{i}", "ₙ": " _{n}", "ₚ": " _{p}",

               "½": "\\text{\\textonehalf}",

               "⌈": "\lceil ", "⌉": "\\rceil ", "⌊": "\lfloor ", "⌋": "\\rfloor ",
               "□": "\square ", "◇": "\diamond ", "○": "\circ ",
               
               "∗": "*", "•": "\\bullet ", "∙": "\cdot ", "⋅": "\cdot ", "⋯": "\cdots ",
               "×": "\\times ",
               chr(160): " ", chr(770): "\hat{ }", chr(8198): " ", chr(8201): " ", chr(8202): " ", chr(8239): " ", chr(65024): " ",
               }  # define desired replacements here

        # use these three lines to do the replacement
        rep = dict((re.escape(k), v) for k, v in rep.items())
        pattern = re.compile("|".join(rep.keys()))
        latex_string = pattern.sub(lambda m: rep[re.escape(m.group(0))], latex_string)

        return latex_string

    @staticmethod
    def replace_invalid_commands(latex_string):
        modified_latex_string = latex_string

        if r"\and" in modified_latex_string:
            match = LaTeX_Helper.REAndCommand.search(modified_latex_string, 0)
            while match:
                start, end = match.span()

                if end - start == 4:
                    # exact match ..., do replace ...
                    prefix = modified_latex_string[:start]
                    middle = r"\land"
                    suffix = modified_latex_string[end:]
    
                    modified_latex_string = prefix + middle + suffix

                # find next (if any)
                match = LaTeX_Helper.REAndCommand.search(modified_latex_string, start + 1)

            # modified_latex_string = modified_latex_string.replace("\\and ", "\land ")

        match = LaTeX_Helper.RESymbolCommand.search(modified_latex_string, 0)
        while match:
            start, end = match.span()

            prefix = modified_latex_string[:start]
            middle = modified_latex_string[start:end - 6]
            suffix = modified_latex_string[end:]

            modified_latex_string = prefix + middle + suffix

            match = LaTeX_Helper.RESymbolCommand.search(modified_latex_string, start + 1)

        if "\par " in modified_latex_string:
            modified_latex_string = modified_latex_string.replace("\par ", " ")

        if "\\xb " in modified_latex_string:
            modified_latex_string = modified_latex_string.replace("\\xb ", " ")

        if "\\u{" in modified_latex_string:
            # place within text environment ...
            next_pos = modified_latex_string.find("\\u{", 0)
            while next_pos >= 0:
                closing_bracket = modified_latex_string.find("}", next_pos + 1)

                # add enclosing command ....
                prefix = modified_latex_string[:next_pos]
                middle = "\\text{" + modified_latex_string[next_pos:closing_bracket + 1] + "}"
                suffix = modified_latex_string[closing_bracket + 1:]

                modified_latex_string = prefix + middle + suffix

                next_pos = modified_latex_string.find("\\u{", next_pos + len(middle))

        # get rid of source coloring ...
        if "\color{" in modified_latex_string:
            next_pos = modified_latex_string.find("\color{", 0)
            while next_pos >= 0:
                closing_bracket = modified_latex_string.find("}", next_pos + 1)

                prefix = modified_latex_string[:next_pos]
                suffix = modified_latex_string[closing_bracket + 1:]

                modified_latex_string = prefix + suffix

                next_pos = modified_latex_string.find("\color{", next_pos)

        return modified_latex_string

    @staticmethod
    def remove_unescaped_symbol(latex_string, symbol):
        pos_next = latex_string.find(symbol, 0)
        while pos_next >= 0:
            if pos_next == 0:
                # invalid ... remove
                latex_string = latex_string[1:]
                pos_next = latex_string.find(symbol, 0)
            else:
                # check previous character ...
                if latex_string[pos_next - 1:pos_next + 1] != '\\' + symbol:
                    # percentage not preceded by \ ... assume error and remove ...
                    latex_string = latex_string[:pos_next] + latex_string[pos_next + 1:]
                    # re-start search from same position
                    pos_next = latex_string.find(symbol, pos_next)
                else:
                    # assume valid ... re-start from next position
                    pos_next = latex_string.find(symbol, pos_next + 1)

        return latex_string

    @staticmethod
    def remove_invalid_spaces(latex_string):
        # remove spaces in places where they are invalid (enviroment names, etc)
        check_commands = ["\\begin", "\end"]
        for command in check_commands:
            pos_next = latex_string.find(command, 0)
            while pos_next >= 0:
                # check for enclosing curly brackets ...
                left_pos = latex_string.find("{", pos_next + 1)
                right_pos = latex_string.find("}", left_pos)

                if 0 < left_pos < right_pos:
                    if " " in latex_string[left_pos + 1:right_pos]:
                        # invalid spaces found ... replace them ...
                        prefix = latex_string[:left_pos + 1]
                        middle = latex_string[left_pos + 1:right_pos].replace(" ", "")
                        suffix = latex_string[right_pos:]

                        latex_string = prefix + middle + suffix

                pos_next = latex_string.find(command, pos_next + 1)

        return latex_string

    @staticmethod
    def generate_rendering_preamble():
        # prepare tex pre-amble ...
        actual_packages = ('dsfont', 'amssymb', 'xcolor', 'cancel', "amsmath", "amsfonts", "arevmath", "txfonts",
                           'eurosym', "mathtools", "textcomp")
        package_includes = "\n" + "\n".join(["\\usepackage{" + p + "}" for p in actual_packages])

        digits = [r"\newcommand{\zero}{0}"]

        english_chars = [r"\newcommand{\B}{\mathbb{B}}", r"\newcommand{\C}{\mathbb{C}}", r"\newcommand{\K}{\mathbb{K}}",
                         r"\newcommand{\R}{\mathbb{R}}", r"\newcommand{\Q}{\mathbb{Q}}", r"\newcommand{\Z}{\mathbb{Z}}"]

        greek_chars = [r"\newcommand{\Alpha}{A}", r"\newcommand{\Eta}{E}", r"\newcommand{\Epsilon}{E}",
                       r"\newcommand{\Iota}{\iota}", r"\newcommand{\Mu}{\mathrm{M}}", r"\newcommand{\Nu}{\mathrm{N}}",
                       r"\newcommand{\Omicron}{\mathrm{O}}", r"\newcommand{\omicron}{\mathrm{o}}",
                       r"\newcommand{\PI}{\Pi}", r"\newcommand{\Rho}{\rho}", r"\newcommand{\Tau}{\tau}",
                       r"\newcommand{\Chi}{X}"]

        num_sets = [r"\newcommand{\real}{\mathbb{R}}", r"\newcommand{\reals}{\mathbb{R}}",
                    r"\newcommand{\Complex}{\mathbb{C}}", r"\newcommand{\Zeta}{\mathbb{Z}}"]

        alt_commands = [r"\newcommand{\alef}{\aleph}", r"\newcommand{\ang}{\angle}", r"\newcommand{\bull}{\bullet}",
                        r"\newcommand{\darr}{\downarrow}", r"\newcommand{\Dagger}{\dagger}",
                        r"\newcommand{\exist}{\exists}", r"\newcommand{\isin}{\in}", r"\newcommand{\Larr}{\Leftarrow}",
                        r"\newcommand{\lrarr}{\leftrightarrow}", r"\newcommand{\plusmn}{\pm}",
                        r"\newcommand{\Rarr}{\Rightarrow}", r"\newcommand{\sdot}{\cdot}",
                        r"\newcommand{\sub}{\subset}", r"\newcommand{\sube}{\subseteq}",
                        r"\newcommand{\supe}{\supseteq}", r"\newcommand{\uarr}{\uparrow}"]

        operators = [r"\DeclareMathOperator{\arcsec}{arcsec}", r"\DeclareMathOperator{\arccot}{arccot}",
                     r"\DeclareMathOperator{\arccsc}{arccsc}"]

        custom_commands = digits + english_chars + greek_chars + num_sets + alt_commands + operators

        command_definitions = "\n" + "\n".join(custom_commands)

        preamble = r"\documentclass[12pt]{article}" + "\n"
        preamble += "\pagestyle{empty}\n"
        preamble += package_includes
        preamble += command_definitions
        preamble += "\n" + r"\begin{document}"

        return preamble

    @staticmethod
    def try_render_string(latex_string, output_file, dviopt, ext):
        # generate static preamble only once
        if LaTeX_Helper.RenderPreamble is None:
            LaTeX_Helper.RenderPreamble = LaTeX_Helper.generate_rendering_preamble()

        try:
            # render ....
            sympy.preview(latex_string, viewer="file", filename=output_file, euler=False, dvioptions=dviopt,
                          preamble=LaTeX_Helper.RenderPreamble, output=ext)

            return True
        except Exception as e:
            print(e)
            return False

    @staticmethod
    def prepare_latex_string(latex_string):
        # check if string has a math environment ...
        if latex_string[0] != "$" and latex_string[:2] != "\[":
            # does not have, check type to use ...
            if "\\begin" in latex_string:
                # handle as special environment
                # use \[ \] instead of $ ... $
                modified_latex_string = "\[\n" + latex_string + " \n\]\n"
            else:
                # inline mode ...
                modified_latex_string = "$ " + latex_string + " $"
        else:
            modified_latex_string = latex_string

        # remove known invalid characters ... (&gt; -> >)
        modified_latex_string = LaTeX_Helper.replace_invalid_chars(modified_latex_string)

        # a + % x -> a + x
        if "%" in modified_latex_string:
            modified_latex_string = LaTeX_Helper.remove_unescaped_symbol(modified_latex_string, "%")

        if "#" in modified_latex_string:
            modified_latex_string = LaTeX_Helper.remove_unescaped_symbol(modified_latex_string, "#")

        # \exist -> exists , ...
        modified_latex_string = LaTeX_Helper.replace_invalid_commands(modified_latex_string)

        # \end{ array} -> \end{array}
        modified_latex_string = LaTeX_Helper.remove_invalid_spaces(modified_latex_string)

        return  modified_latex_string

    @staticmethod
    def render(latex_string, output_file, dvi=600, verbose=True):
        if verbose:
            print("Rendering {0:s} to {1:s}".format(latex_string, output_file))

        base, ext = os.path.splitext(output_file)
        ext = ext[1:]
        if ext == "png":
            dviopt = ["-D", str(dvi), "-T", "tight", "-z", "9", "--truecolor"]
        else:
            dviopt = []

        modified_latex_string = LaTeX_Helper.prepare_latex_string(latex_string)

        # try rendering ....
        if LaTeX_Helper.try_render_string(modified_latex_string, output_file, dviopt, ext):
            return latex_string, True

        return latex_string, False

    @staticmethod
    def replace_qvars(render_string, char_replace, query_var_prefix="\qvar"):
        # find and replace query variables ....
        qvars_replaced = 0
        while query_var_prefix in render_string:
            # find location of query variable in input string
            next_qvar_pos = render_string.find(query_var_prefix)
            end_next_qvar = render_string.find("}", next_qvar_pos)

            prefix = render_string[:next_qvar_pos]
            suffix = render_string[end_next_qvar + 1:]

            render_string = prefix + "\\textcolor{red}{" + char_replace + "}" + suffix
            qvars_replaced += 1

        return render_string, qvars_replaced

    @staticmethod
    def rec_latex_empty_frac(depth):
        if depth == 0:
            return "~"
        else:
            inner = LaTeX_Helper.rec_latex_empty_frac(depth - 1)
            return "\\frac{" + inner + "}{" + inner + "}"

    @staticmethod
    def extract_leftmost_symbol(symbol_img):
        current_cc = RecognitionHelper.CCsFromBinary(symbol_img)

        all_ccs = [(cc.min_x, cc.min_y, cc) for cc in current_cc]
        all_ccs = sorted(all_ccs, key=lambda x: (x[0], x[1]))

        first_cc = all_ccs[0][2]

        return first_cc.img

    @staticmethod
    def render_latex_labels(latex_labels, mapping_label_latex, vertical_sizes, horizontal_sizes, dvi=300,
                            temporal_filename="DELETE_TEMPORAL_LATEX_IMG.png"):
        latex_images = {}

        for label in latex_labels:
            label_latex_strings = []
            multi_size_render = False

            if label == "\sqrt":
                multi_size_render = True
                for vert_size in vertical_sizes:
                    base_filler = LaTeX_Helper.rec_latex_empty_frac(vert_size)

                    for hor_size in horizontal_sizes:
                        latex_filler = base_filler * hor_size

                        latex_str = "\sqrt{" + latex_filler + "}"
                        label_latex_strings.append(latex_str)
            elif label in ["\{", "\}", "(", ")", "[", "]"]:
                # brackets ...
                multi_size_render = True

                for vert_size in vertical_sizes:
                    base_filler = LaTeX_Helper.rec_latex_empty_frac(vert_size)

                    latex_str = "\left" + label + base_filler + " \\right" + label
                    label_latex_strings.append(latex_str)
            else:
                # draw single size, multiple styles ...
                draw_styles = ["\mathrm", "\mathsf", "\mathtt", "\mathbf", "\mathit"]

                if "A" <= label <= "Z":
                    draw_styles.append("\mathcal")
                    draw_styles.append("\mathds")

                # add default normal style
                label_latex_strings.append(mapping_label_latex[label])

                # add all other styles
                for style in draw_styles:
                    latex_str = style + "{" + mapping_label_latex[label] + "}"
                    label_latex_strings.append(latex_str)

            # from Latex to image ...
            latex_images[label] = []
            for ltx_string in label_latex_strings:
                # render to file ...
                LaTeX_Helper.render("$" + ltx_string + "$", temporal_filename, dvi, False)

                # read back ...
                symbol_img = cv2.imread(temporal_filename, cv2.IMREAD_GRAYSCALE)
                # Delete temporary file ...
                os.remove(temporal_filename)

                # invert image and only keep one channel ...
                symbol_img = 255 - symbol_img

                if multi_size_render:
                    symbol_img = LaTeX_Helper.extract_leftmost_symbol(symbol_img)

                latex_images[label].append(symbol_img)

        return latex_images

    @staticmethod
    def expand_image_sets_per_label(latex_images, min_class_count, max_size_dist):
        # use a simple distortion model to create synthetic copies of the latex symbols. For now, simply
        # apply horizontal/vertical stretching...

        final_per_label = {}
        for label in latex_images:
            final_per_label[label] = []

            n_label_images = len(latex_images[label])
            for idx in range(max(min_class_count, n_label_images)):
                base_image = latex_images[label][idx % n_label_images]
                if idx < n_label_images:
                    final_per_label[label].append(base_image)
                else:
                    dist_dir = random.random()
                    dist_value = random.random() * 2.0 - 1.0

                    new_h, new_w = base_image.shape
                    if dist_dir > 0.5:
                        # horizontal distortion ...
                        new_w = int(new_w * (1 + dist_value * max_size_dist))
                    else:
                        # vertical distortion ...
                        new_h = int(new_h * (1 + dist_value * max_size_dist))

                    distorted_image = cv2.resize(base_image, (new_w, new_h))

                    final_per_label[label].append(distorted_image)

        return final_per_label