
//	gcc -m64 -shared tangentv_lib.c -o tangentv_lib.so

#include <stdio.h>
#include <stdlib.h>

int CC_Boundaries(int* labels, int width, int height, int count_labels,
                  int* out_mins_y, int* out_maxs_y, int* out_mins_x, int* out_maxs_x,
                  int* out_counts)
{

    //Initialize output ...
    int i;
    for (i = 0;  i < count_labels; i++)
    {
        out_mins_y[i] = height;
        out_maxs_y[i] = 0;

        out_mins_x[i] = width;
        out_maxs_x[i] = 0;

        out_counts[i] = 0;
    }

    int x, y;
    int idx = 0;
    for (y = 0; y < height; y++)
    {
        for (x = 0; x < width; x++)
        {
            if (labels[idx] > 0){
                int cc_id = labels[idx] - 1;

                if (out_mins_y[cc_id] > y){
                    out_mins_y[cc_id]  = y;
                }

                if (out_maxs_y[cc_id] < y){
                    out_maxs_y[cc_id] = y;
                }

                if (out_mins_x[cc_id] > x){
                    out_mins_x[cc_id]  = x;
                }

                if (out_maxs_x[cc_id] < x){
                    out_maxs_x[cc_id] = x;
                }

                out_counts[cc_id]++;
            }
            idx ++;
        }
    }

    return 0;
}

// Greedily chose 1-to-1 matching pairs from a scored list of pairs
/*
    double* data: pointer to data array containing all pairs with query pair id, img pair id, score and other
                  size: n_pairs x n_fields

    int n_q_pairs, int n_c_pairs: Are the number of pairs in the original query and candidate graphs

    int q_pair_id_idx, int c_pair_id_idx,  int score_idx:  field indices ....

    int* out_matching: output array where matching pairs will be marked with 1, and everything else with 0

*/
double Greedy_Pair_Matching(double* data, int n_pairs, int n_fields, int q_pair_id_idx, int c_pair_id_idx,
                            int score_idx, int* out_matching){
    int pair_idx;
    int pair_offset = 0;
    int query_pair_id;
    int unique_pair_id;
    int count_pairs_matched = 0;
    double pair_score;
    double total_score = 0.0;
    int* q_pair_used;
    int* c_pair_used;

    if (n_pairs <= 0){
        // Nothing to do really ...
        return total_score;
    }

    // Determine range for query and candidate pairs ids
    int min_q_pair_id = (int)data[q_pair_id_idx];
    int max_q_pair_id = (int)data[q_pair_id_idx];
    int min_c_pair_id = (int)data[c_pair_id_idx];
    int max_c_pair_id = (int)data[c_pair_id_idx];

    //printf("1");
    // Do a first pass ....
    for (pair_idx = 0;  pair_idx < n_pairs; pair_idx++){
        query_pair_id = (int)(data[pair_offset + q_pair_id_idx]);
        unique_pair_id = (int)(data[pair_offset + c_pair_id_idx]);

        // Check for new min/max for query pair id
        if (query_pair_id < min_q_pair_id){
            //New min
            min_q_pair_id = query_pair_id;
        } else {
            if (query_pair_id > max_q_pair_id){
                //New max
                max_q_pair_id = query_pair_id;
            }
        }

        // Check for new min/max for candidate pair id
        if (unique_pair_id < min_c_pair_id){
            //New min
            min_c_pair_id = unique_pair_id;
        } else {
            if (unique_pair_id > max_c_pair_id){
                //New max
                max_c_pair_id = unique_pair_id;
            }
        }

        pair_offset += n_fields;
    }

    //printf("... 2");

    int n_q_pairs = max_q_pair_id - min_q_pair_id + 1;
    int n_c_pairs = max_c_pair_id - min_c_pair_id + 1;
    // printf("Query: [%d, %d]=%d, Candidate: [%d, %d]=%d \n", min_q_pair_id, max_q_pair_id, n_q_pairs, min_c_pair_id, max_c_pair_id, n_c_pairs);

    // Allocating arrays ...
    q_pair_used = malloc(sizeof(int) * n_q_pairs);
    c_pair_used = malloc(sizeof(int) * n_c_pairs);

    //printf("... 3");

    //Initializing ... (marking elements as unused)
    for (pair_idx = 0;  pair_idx < n_q_pairs; pair_idx++){
        q_pair_used[pair_idx] = -1;
    }
    for (pair_idx = 0;  pair_idx < n_c_pairs; pair_idx++){
        c_pair_used[pair_idx] = -1;
    }

    // printf("... 4");

    // For each pair in the array ...
    int q_id_offset;
    int c_id_offset;

    pair_offset = 0;
    for (pair_idx = 0;  pair_idx < n_pairs; pair_idx++){
        query_pair_id = (int)(data[pair_offset + q_pair_id_idx]);
        unique_pair_id = (int)(data[pair_offset + c_pair_id_idx]);
        pair_score = (data[pair_offset + score_idx]);

        //printf("%d, %d, %d, %d, %f\n", pair_idx, pair_offset, query_pair_id, unique_pair_id, pair_score);

        // check if both the query and the candidate pair have not been assigned yet ...
        q_id_offset = query_pair_id - min_q_pair_id;
        c_id_offset = unique_pair_id - min_c_pair_id;
        if (q_pair_used[q_id_offset] == -1 && c_pair_used[c_id_offset] == -1){
            // mark these elements as matched
            q_pair_used[q_id_offset] = unique_pair_id;
            c_pair_used[c_id_offset] = query_pair_id;

            //add the score ...
            total_score += pair_score;

            // mark as matched in the output array (which is assumed to be initialized to 0
            out_matching[pair_idx] = 1;

            // Check for early termination ...
            count_pairs_matched += 1;
            if (count_pairs_matched >= n_q_pairs || count_pairs_matched >= n_c_pairs){
                break;
            }
        }

        // Move to the logical position of the next
        pair_offset += n_fields;
    }

    //printf("... 5");

    // releasing arrays ...
    free(q_pair_used);
    free(c_pair_used);

    //printf("... 6\n");

    return total_score;
}

