
import os
import sys
import pickle
import codecs

from AM_CommonTools.configuration.configuration import Configuration
from TangentV.indexing.index_parameters import IndexParameters
from TangentV.util.mathscraper_helper import MathScraperHelper

if sys.stdout.encoding != 'utf8':
    sys.stdout = codecs.getwriter('utf8')(sys.stdout.buffer, 'strict')
if sys.stderr.encoding != 'utf8':
    sys.stderr = codecs.getwriter('utf8')(sys.stderr.buffer, 'strict')


def main():
    if len(sys.argv) < 2:
        print("Usage:")
        print("python {0:s} tanv_conf [workers]".format(sys.argv[0]))
        print("\nWhere:")
        print("\ttanv_conf:\tTangent-V Configuration File")
        print("\tworkers:\tNumber of parallel processes used")
        return

    tanv_config_filename = sys.argv[1]

    try:
        tanv_config = Configuration.from_file(tanv_config_filename)
        index_params = IndexParameters.ReadFromConfiguration(tanv_config)
    except:
        print("Invalid Tangent-v configuration")
        return

    # input_directory = sys.argv[1]
    output_filename = tanv_config.get("TEMPORARY_PDF_SYMBOLS")

    if len(sys.argv) >= 3:
        try:
            batch_workers = int(sys.argv[2])
        except:
            print("Invalid number of workers")
            return
    else:
        batch_workers = 1

    input_directory = tanv_config.get("BINARY_IMAGES_DIRECTORY")
    math_scrapper_path = tanv_config.get("MATH_SCRAPPER_PATH")

    math_scrapper = MathScraperHelper(math_scrapper_path)
    all_data, _ = math_scrapper.extract_data(input_directory, batch_workers, True)

    with open(output_filename, "wb") as output_file:
        pickle.dump(all_data, output_file, pickle.HIGHEST_PROTOCOL)

    print("Process completed")

if __name__ == "__main__":
    main()

